;; Generated package description from forgejo.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "forgejo" "0.1.2" "Emacs Forgejo Front-end" '((emacs "29.1") (markdown-mode "2.6") (keymap-popup "0.2.1")) :commit "0ab81526cef90e8fdfa91a1c98f72c8168335fca" :authors '(("Thanos Apollo" . "public@thanosapollo.org")) :maintainer '("Thanos Apollo" . "public@thanosapollo.org") :keywords '("tools" "vc" "git" "forgejo") :url "https://codeberg.org/thanosapollo/emacs-forgejo")
